<?php

namespace Polkurier;

use Polkurier\Exception\ApiException;

/**
 * Class ApiClient
 * @package Polkurier
 */
class ApiClient
{

    /**
     * @var string
     */
    protected $apiLogin;

    /**
     * @var string
     */
    protected $apiToken;

    /**
     * @var string
     */
    protected $apiUrl = "https://api.polkurier.pl";

    /**
     * @var bool string
     */
    public $debug = false;

    /**
     * @var array
     */
    private $requestCache = [];

    /**
     * @var array|null
     */
    protected static $postOfficesCache;

    /**
     * ApiClient constructor.
     * @param $login
     * @param $token
     */
    public function __construct($login, $token)
    {
        $this->apiLogin = (string)$login;
        $this->apiToken = (string)$token;
    }

    /**
     * @return ApiClient
     */
    public static function getInstance()
    {
        return new self(get_option('polkurier_account_id'), get_option('polkurier_apikey'));
    }

    /**
     * @param $method
     * @param array $params
     * @return array|mixed
     * @throws ApiException
     */
    protected function call($method, $params = array())
    {
        $this->lastError = null;
        $callParams = [
            "authorization" => [
                "login" => $this->apiLogin,
                "token" => $this->apiToken,
            ],
            "apimetod" => $method,
            "data" => $params,
        ];
        $callParamsText = json_encode($callParams, JSON_NUMERIC_CHECK);
        $callParamsText = str_replace('##POLKURIERMAGIC##', '', $callParamsText);
        if ($this->debug) {
            echo "\nREQUEST:\n$callParamsText\n";
        }

        $headers = [
            'Content-Type: application/json',
            'Content-Length: ' . strlen($callParamsText),
        ];
        $ch = curl_init();
        curl_setopt($ch, CURLOPT_URL, $this->apiUrl);
        curl_setopt($ch, CURLOPT_CUSTOMREQUEST, "POST");
        curl_setopt($ch, CURLOPT_POSTFIELDS, $callParamsText);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
        curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, false);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
        $result = curl_exec($ch);
        $http_code = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        curl_close($ch);

        if ($http_code !== 200) {
            throw new ApiException("Błąd połączenia z API serwisu POLKURIER.PL. CODE $http_code (E_NO_CONNECTION)");
        }

        $arrayResult = json_decode($result, true);
        if ($this->debug) {
            echo "\nRESPONSE:\n$";
            print_r($arrayResult);
            echo "\n";
        }
        if (isset($arrayResult["status"]) && $arrayResult["status"] === 'error') {
            throw new ApiException($arrayResult["response"]);
        }

        if (!is_array($arrayResult)) {
            throw new ApiException('Błąd połączenia z API serwiesu POLKURIER.PL. (E_EMPTY_RESPONSE)');
        }
        return $arrayResult;
    }


    /**
     * Wykonuje zapytanie do API
     * @param string $method
     * @param array $params
     * @return array
     * @throws ApiException
     */
    public function makeApiRequest($method, array $params)
    {
        return $this->call((string)$method, $params)['response'];
    }


    /**
     * @param bool $additionalData
     * @return array|mixed
     * @throws ApiException
     */
    public function getAvailableCarriers($additionalData = false)
    {
        $params = [
            'additional_data' => $additionalData,
        ];

        $cacheKey = md5('available_carriers' . serialize($params));
        if (isset($this->requestCache[$cacheKey])) {
            return $this->requestCache[$cacheKey];
        }

        $data = $this->call("available_carriers", $params);
        if (isset($data["response"])) {
            $this->requestCache[$cacheKey] = $data["response"];
            return $data["response"];
        }
        return array();
    }

    /**
     * @return array|mixed
     * @throws ApiException
     */
    public function getCouriersMessages()
    {
        $cacheKey = md5('available_carriers');
        if (isset($this->requestCache[$cacheKey])) {
            return $this->requestCache[$cacheKey];
        }

        $data = $this->call("get_couriers_messages");
        if (isset($data["response"])) {
            $this->requestCache[$cacheKey] = $data["response"];
            return $data["response"];
        }
        return array();
    }

    /**
     * @param $package
     * @return array|mixed
     * @throws ApiException
     */
    public function orderValuation($package)
    {
        $type = $package->getType();
        $data = [[
            "length" => $package->depth,
            "width" => $package->width,
            "height" => $package->height,
            "weight" => $package->weight,
            "amount" => $package->count,
            "type" => $package->nonstandard ? "NST" : "ST",
        ]];
        $params = [
            "shipmenttype" => $type,
            "packs" => $data,
        ];
        $data = $this->call("order_valuation", $params);
        if (isset($data["response"])) {
            return $data["response"];
        } else {
            return array();
        }

    }

    /**
     * @param $courier
     * @param $packages
     * @param $sender
     * @param $recipient
     * @param $insurance
     * @param $pickup
     * @param $description
     * @param $pobranie
     * @param $courierservice
     * @param $orderDetails
     * @return array|mixed
     * @throws ApiException
     */
    public function createOrder($courier, $packages, $sender, $recipient, $insurance, $pickup, $description, $pobranie, $courierservice, &$orderDetails)
    {
        $packageData = array();
        foreach ($packages as $package) {
            $packageData[] = [
                "length" => $package->depth,
                "width" => $package->width,
                "height" => $package->height,
                "weight" => $package->weight,
                "amount" => $package->count,
                "type" => $package->nonstandard ? "NST" : "ST",
            ];
        }
        if ($sender->address["account_id"] && $courier === 'GEIS') {
            $sender->address["courieraccount"] = true;
            $sender->address["addressbookid"] = $sender->address["account_id"];
        }
        unset($sender->address["account_id"]);
        $params = [
            "platform" => 2,
            "courier" => $courier,
            "shipmenttype" => $package->getType(),
            "description" => $description,
            "sender" => $sender->address,
            "recipient" => $recipient->address,
            "packs" => $packageData,
            "pickup" => $pickup,
            "insurance" => $insurance,
            'courierservice' => (array)$courierservice,
        ];
        if ($pobranie) {
            $params["COD"] = $pobranie;
        }
        $orderDetails = $params;
        $data = $this->call("create_order", $params);
        if (isset($data["response"])) {
            return $data["response"];
        } else {
            return array();
        }

    }

    /**
     * @param bool $cod
     * @param bool $send
     * @return array|mixed
     * @throws ApiException
     */
    public function getInpost($cod = false, $send = false)
    {
        $params = [
            "cod_available" => $cod,
            "parcel_send" => $send,
        ];
        $data = $this->call("inpost_parcel_machines", $params);
        if (isset($data["response"])) {
            return $data["response"];
        } else {
            return array();
        }

    }

    /**
     * @return mixed
     * @throws ApiException
     */
    public function getPostOffices()
    {
        if (self::$postOfficesCache === null) {
            $data = $this->call('pocztex_post_offices');
            if (isset($data["response"])) {
                self::$postOfficesCache = $data["response"];
            } else {
                self::$postOfficesCache = array();
            }
        }
        return self::$postOfficesCache;
    }

    /**
     * @param $statusSymbol
     * @return string
     */
    public static function getStatusText($statusSymbol)
    {
        switch ($statusSymbol) {
            case "O":
                return "oczekuje";
            case "P":
                return "potwierdzone";
            case "A":
                return "anulowane";
            case "WP":
                return "w przewozie";
            case "D":
                return "dostarczona";
            case "Z":
                return "zwrot do nadawcy";
            case "W":
                return "wyjątek";
        }
        return '';
    }

    /**
     * @param $statusSymbol
     * @return string
     */
    public static function getStatusTextExtended($statusSymbol)
    {
        switch ($statusSymbol) {
            case "O":
                return "Zlecenie zostało prawidłowo zapisane i oczekuje na płatność.";
            case "P":
                return "Zlecenie zostało prawidłowo zapisane. Wygenerowano list przewozowy. Przesyłka oczekuje na odbiór przez kuriera.";
            case "A":
                return "Przesyłka anulowana.";
            case "WP":
                return "Przesyłka odebrana od nadawcy przez kuriera i jest w drodze do adresata.";
            case "D":
                return "Przesyłka dotarła do adresata.";
            case "Z":
                return "Odbiorca odmówił odebrania przesyłki która zostanie zwrócona nadawcy.";
            case "W":
                return "Status informuje że pojawiły się problemy z dostarczeniem przesyłki. W przypadku wystąpienia tego statusu zalecamy kontakt z BOK w celu szybkiego wyjaśnienia z przewoźnikiem.";
        }
        return '';
    }


    /**
     * @param $packageId
     * @return array|mixed
     * @throws ApiException
     */
    public function getStatus($packageId)
    {
        $params = [
            "orderno" => $packageId,
        ];
        $data = $this->call("get_status", $params);
        if (isset($data["response"])) {
            return $data["response"];
        }
        return array();
    }

    /**
     * @param $packageId
     * @return array|mixed
     * @throws ApiException
     */
    public function cancelOrder($packageId)
    {
        $params = [
            "orderno" => $packageId,
        ];
        $data = $this->call("cancel_order", $params);
        if (isset($data["response"])) {
            return $data["response"];
        }
        return array();
    }

    /**
     * @param $pickupdate
     * @param $courier
     * @param $shipfrom
     * @param $parcel
     * @return array|mixed
     * @throws ApiException
     */
    public function pickupCourier($pickupdate, $courier, $shipfrom, $parcel)
    {
        $params = [
            "pickupdate" => $pickupdate,
            "courier" => $courier,
            "shipfrom" => $shipfrom,
            "parcel" => $parcel,
        ];
        $data = $this->call("pickup_courier", $params);
        if (isset($data["response"])) {
            return $data["response"];
        }
        return array();
    }

    /**
     * @param $orderId
     * @return false|string|null
     * @throws ApiException
     */
    public function getLabel($orderId)
    {
        if (!is_array($orderId)) {
            $params = [
                "orderno" => [$orderId],
            ];
        } else {
            $params = [
                "orderno" => $orderId,
            ];
        }
        $data = $this->call("get_label", $params);
        if (isset($data["status"]) && $data["status"] === "success") {
            if (!empty($data["response"]["file"])) {
                return base64_decode($data["response"]["file"]);
            }

        } else {
            return null;
        }
        return null;
    }


    /**
     * @return string|null
     * @throws ApiException
     */
    public function getMapToken()
    {
        $data = $this->call('get_map_token');
        if (isset($data['response']['token'])) {
            return (string)$data['response']['token'];
        }
        return null;
    }


    /**
     * @param array $couriers
     * @param string $searchQuery
     * @param bool $withCod
     * @param bool $withCollect
     * @param bool $withSend
     * @return array
     * @throws ApiException
     */
    public function getCourierPoint($couriers = [], $searchQuery = '', $withCod = false, $withCollect = false, $withSend = false)
    {
        $functions = [];
        if ($withCod) {
            $functions[] = 'cod';
        }
        if ($withCollect) {
            $functions[] = 'collect';
        }
        if ($withSend) {
            $functions[] = 'send';
        }
        $data = $this->call('get_courier_point', [
            'couriers' => $couriers,
            'searchquery' => $searchQuery,
            'functions' => $functions
        ]);
        return (array)$data['response'];
    }

}
